/* The same as Tree but with bounding box method
for drawing trees */
import java.awt.*;
public class BoundBoxedTree extends Tree
{
	private int width;	//width of tree in pixels
	private int depth;	//depth of tree in levels of nodes							
	private static int hDistance = 10; //minimal allowed horisontal
	private static int vDistance =20; //and vertical distance between nodes
	private static int initvDistance =20; //and vertical distance between nodes
	protected  Point corner = new Point(10,10); //left upper corner
	private Point bestPlace;//optimal place for the center of the root node
	private MovingPoint bestLine;


	BoundBoxedTree()
	{
		this(new Node());
		bestFit(corner.x,corner.y);
	}

	BoundBoxedTree(BoundBoxedTree left,Node middle,BoundBoxedTree right)
	{
		super(left,middle,right);
 		width = 2*root.getRadius();
		if (left != null)
			width += hDistance + left.width;
		if (right != null)
			width += hDistance + right.width;

		this.depth = 1 + Math.max ( (left != null ? left.depth : 0),
			(right != null ?right.depth :0));
		bestFit(corner.x,corner.y);
	}

	BoundBoxedTree(Node node)
	{
		super(node);
		this.width = 2 * root.getRadius();
		this.depth = 1;
		bestFit(corner.x,corner.y);
	}

	synchronized void setVdistance(int arg)
	{
		vDistance = arg;
		updateInfo();
		bestFit(corner.x,corner.y);
	}



	public synchronized int getVdistance()
	{
		return( vDistance);
	}

	public int getHdistance()
	{
		return ( hDistance);
	}

	public int getDepth()
	{
		return ( depth );
	}

	public int getWidth()
	{
		return ( width + 60 );  //returns widths of the tree in pixels plus some extra
								//space for scrollbars
	}

	public int getHeight()	 //height of the tree in pixels
	{
		return ( (1+depth)*(2*root.getRadius() + vDistance) );
	}



	public synchronized void updateInfo()
	{	try{
		int left_width,right_width,left_depth,right_depth;
		if (left == null)
		{
			left_depth = 0;
			left_width = -hDistance;//used to compute width uniformly-always add
								//2*hDistance, even in the case of empty sibling	
		}
		else 
		{
			((BoundBoxedTree)left).updateInfo();
			left_width = ((BoundBoxedTree)left).width;
			left_depth = ((BoundBoxedTree)left).depth;
		}

		if (right == null)
		{
			right_depth = 0;
			right_width = -hDistance;
		}
		else 
		{
			((BoundBoxedTree)right).updateInfo();
			right_width = ((BoundBoxedTree)right).width;
			right_depth = ((BoundBoxedTree)right).depth;
		}

		width = left_width + right_width + 2*hDistance + 2*root.getRadius();
		depth = Math.max ( left_depth , right_depth ) + 1;
	}	 catch (Exception e) {System.out.println("BBupdateError");};
	}

	public synchronized void bestFit(int lCornerX,int lCornerY)
	{
		try{
		int x,y;  //optimal location for left corner of the node
		int dx,
			dy = 2*root.getRadius() + vDistance;;
		int lw;
		int r = root.getRadius();

		if ( left != null)
		{
		 	x = 
			   lCornerX + ((BoundBoxedTree)left).width + hDistance + r ;
			y = lCornerY;
			dx = 0;
			((BoundBoxedTree)left).bestFit(lCornerX + dx,lCornerY + dy);
		}
		else
		{
			//!!!x and y in node are left corner, not the center.Change!!!
			x = lCornerX + r;
			y = lCornerY;
		}

		if ( right != null)
		{
			dx =  x + r + hDistance - lCornerX;
			((BoundBoxedTree)right).bestFit(lCornerX + dx,lCornerY + dy);
		}
		bestPlace = new Point(x, y) ;
		bestPlace.translate(0,r) ;
		bestLine = new MovingPoint(root.center(),bestPlace);

	   	}catch (Exception e) {System.out.println("bestFitError"+ e.toString());};

	}

	public synchronized void fine_draw(Graphics g)
	{
		updateInfo();
		bestFit(corner.x,corner.y);
		draw(g);
	}
	
	private synchronized boolean basic_step()
	{
		if (bestLine.isDone())
			return false;            //the jorney is over
		root.center(bestLine.step());
		return true;          //should step more
	}

	public synchronized boolean step()
	{
		boolean value1 = false , value2 = false ,value;	  
		value = basic_step();
		if(left != null)
			value1 = ((BoundBoxedTree)left).step();
		if(right != null)
			value2 = ((BoundBoxedTree)right).step();
		return (value || value1 || value2); //if false - the work is done 
	}

}


													 