import java.awt.*;
class  Splay  
{
	/*search tree. The keys are sorted from left - the least to right the greatest */
	public BoundBoxedTree splay;	 //this tree may be emtpy
	public String CurrentMessage;	 //message passed for printing to MyCanvas
	Splay(Node node)
	{
		splay = new BoundBoxedTree(node);
	}

	Splay ()
	{
		splay =null;
	}

	Splay (BoundBoxedTree tree)
	{
		splay = tree;
	}		  

	public int getHeight()
	{
		return	(splay != null ? splay.getHeight():0);
	}
	public int getWidth()
	{
		return(splay != null ? splay.getWidth() : 0);
	}

	public synchronized Tree locate(int key)
	{
		return(locate (key,splay));
	}

	public synchronized Tree locate(int key,Tree tree )
	{
		boolean found = false;
		if (tree == null)
		{
			CurrentMessage = "Empty tree";
			return null;
		}
		int currentKey = tree.root.getValue();
		try 
		{
			/* exit from the while on exception or when the key is not found */
			while (!found)
			{
				if(currentKey > key)
				{
					currentKey = tree.left.root.getValue(); //if tree.left == null - Exception
					tree = tree.left;
					continue;
				}
 				if(currentKey < key)
				{
					currentKey = tree.right.root.getValue();
					tree = tree.right;
					continue;
				}
				//otherwise
				found = true;
			}
		}
		catch(NullPointerException e)
		{
			//O.K. The key isn't found. Exit from the loop with found == false
		}
		if(found == true)
			CurrentMessage = Integer.toString(key)+" found";
		else CurrentMessage = Integer.toString(key)+" not found";
		return tree;
	}

	public synchronized Tree Insert(int key)
	{
		if (splay == null)
		{
			CurrentMessage = Integer.toString(key)+" inserted";
			return(splay = new BoundBoxedTree(new Node(new Point(1,1),key)));
		}
		Tree located = locate(key);
		if(located.root.getValue() == key)
		{
			CurrentMessage = Integer.toString(key)+" was in tree";
			return(located);
		}
		CurrentMessage = Integer.toString(key)+" inserted";
		return(fineAttach((BoundBoxedTree)located,key));
	}

	public synchronized Tree Delete(int key)
	{
		/* returns the father of the node just before deletion.
		If the node "located" with val=key has left son, we wind the leftmost predecessor - leaf,
		swap it (its value) with "located",delete the leaf and ,finaly swap(leaf.father).
		Otherwise we delete "located" and swap from its father*/
		Tree swaped;
		if (splay == null)
		{	
			CurrentMessage = "Empty tree ";
			return null;
		}
		Tree located = locate(key);

		if(located.root.getValue() != key)
		{
			CurrentMessage = Integer.toString(key)+" not found";
			return located;		//key not found
		
		}

		if (located.left != null)		//swap with the predecessor
		{
			swaped = locate(key,located.left);
			located.root.setValue(swaped.root.getValue());
			if(located.left.right != null)
				swaped.father.right = swaped.left;
			else 	located.left = swaped.left;	 
			
			if (swaped.left != null)
				swaped.left.father = swaped.father;
			CurrentMessage = Integer.toString(key)+" deleted";
			return swaped.father;
		}
		else		 // located.left == null
		{
			if (located.father == null)		//located is the root and it doesn't have left son
			{
				splay = (BoundBoxedTree)located.right;
		 		if (located.right != null) 
					located.right.father = null;
				CurrentMessage = Integer.toString(key)+" deleted";
				return(splay);
			}
			
			if (located.father.left == located)
				located.father.left = located.right;
			else  located.father.right = located.right;

			if	 (located.right != null) 
				located.right.father = located.father;
			CurrentMessage = Integer.toString(key)+" deleted";
			return (located.father);
		}
	}


	private synchronized Tree fineAttach(BoundBoxedTree tree,int key)
	{
		/* assumes that the tree doesn't have a son on the side where new node
		should be attached */
		Point center = tree.root.center(); //tree's location
		BoundBoxedTree node;
		if(tree.root.getValue() > key)
		{
			center.translate( -splay.getHdistance()/2 , splay.getVdistance());
			node = new 	 BoundBoxedTree(new Node(center,key));
			tree.left = node;
		}
		else
		{
			center.translate( splay.getHdistance()/2 , splay.getVdistance());
			node = new 	 BoundBoxedTree(new Node(center,key));
			tree.right = node;
		}
		node.father = tree;
		splay.updateInfo();
		splay.bestFit(splay.corner.x,splay.corner.y); 
		return node;
	}


	 public synchronized void ZigLeft(Tree tree)
	 {
		 /* tree is the left son of the root */
		 Tree tempfather = tree.father;
		 splay = (BoundBoxedTree)tree;
		 tree.father = null;
		 tempfather.left = tree.right;

		 tree.right = tempfather;
		 if(tempfather.left != null)
			 tempfather.left.father = tempfather;
		 tempfather.father = tree;
	 }

 	 public synchronized void ZigRight(Tree tree)
	 {
		 /* tree is the right son of the root */
		 Tree tempfather = tree.father;
		 splay = (BoundBoxedTree)tree;
		 tree.father = null;
		 tempfather.right = tree.left;

		 tree.left = tempfather;
		 if(tempfather.right != null)
			tempfather.right.father = tempfather;
		 tempfather.father = tree;
	 }

 	 public synchronized void ZigZigLeft(Tree tree)
	 {
		 /* tree is the left son of the father */
		 Tree grandfather = tree.father.father;
		 Tree parent = tree.father;
		 Tree A,B,C,D; 
		 A = tree.left;
		 B = tree.right;
		 C = parent.right;
		 D = grandfather.right;
		 if(grandfather.father == null)
		 {
			splay = (BoundBoxedTree)tree;
			tree.father = null;
		 }
		 else
		 {
			tree.father = grandfather.father;
			if (grandfather.father.left == grandfather)
				grandfather.father.left = tree;
			else
				grandfather.father.right = tree;
		 }

		 tree.right = parent;
		 parent.right = grandfather;
		 parent.left = B;
		 parent.father = tree;
		 grandfather.left = C;
		 grandfather.father = parent;
		 if ( B != null)
			B.father = parent;
		 if (C != null)
			C.father = grandfather;
	 }

	 public synchronized void ZigZigRight(Tree tree)
	 {
		 /* tree is the left son of the father */
		 Tree grandfather = tree.father.father;
		 Tree parent = tree.father;
		 Tree A,B,C,D; 
		 A = tree.left;
		 B = tree.right;
		 C = parent.left;
		 D = grandfather.left;
		 if(grandfather.father == null)
		 {
			splay = (BoundBoxedTree)tree;
			tree.father = null;
		 }
		 else
		 {
			tree.father = grandfather.father;
			if (grandfather.father.right == grandfather)
				grandfather.father.right = tree;
			else
				grandfather.father.left = tree;
		 }

		 tree.left = parent;
		 parent.left = grandfather;
		 parent.right = A;
		 parent.father = tree;
		 grandfather.right = C;
		 grandfather.father = parent;
		 if ( A != null)
			A.father = parent;
		 if (C != null)
			C.father = grandfather;
	 }

 	 public synchronized void ZigZagLeft(Tree tree)
	 {
		 /* parent is the left son of the grandfather */
		 Tree grandfather = tree.father.father;
		 Tree parent = tree.father;
		 Tree A,B,C,D; 
		 A = parent.left;
		 B = tree.left;
		 C = tree.right;
		 D = grandfather.right;
		 if(grandfather.father == null)
		 {
			splay = (BoundBoxedTree)tree;
			tree.father = null;
		 }
		 else
		 {
			tree.father = grandfather.father;
			if (grandfather.father.left == grandfather)
				grandfather.father.left = tree;
			else
				grandfather.father.right = tree;
		 }

		 tree.left = parent;
		 tree.right = grandfather;
		 parent.left = A;
		 parent.father = tree;
		 parent.right = B;
		 grandfather.left = C;
		 grandfather.right = D;
		 grandfather.father = tree;
		 if ( A != null)
			A.father = parent;
		 if (C != null)
			C.father = grandfather;
		 if ( B != null)
			B.father = parent;
		 if (D != null)
			D.father = grandfather;
	 }

 	 public synchronized void ZigZagRight(Tree tree)
	 {
		 /* parent is the left son of the grandfather */
		 Tree grandfather = tree.father.father;
		 Tree parent = tree.father;
		 Tree A,B,C,D; 
		 A = grandfather.left;
		 B = tree.left;
		 C = tree.right;
		 D = parent.right;
		 if(grandfather.father == null)
		 {
			splay = (BoundBoxedTree)tree;
			tree.father = null;
		 }
		 else
		 {
			tree.father = grandfather.father;
			if (grandfather.father.left == grandfather)
				grandfather.father.left = tree;
			else
				grandfather.father.right = tree;
		 }

		 tree.right = parent;
		 tree.left = grandfather;
		 parent.left = C;
		 parent.father = tree;
		 parent.right = D;
		 grandfather.left = A;
		 grandfather.right = B;
		 grandfather.father = tree;
		 if ( A != null)
			A.father = grandfather;
		 if (B != null)
			B.father = grandfather;
		 if ( C != null)
			C.father = parent;
		 if (D != null)
			D.father = parent;
	 }

	 public synchronized boolean SplayStep(Tree tree)
	 {
	 /*Interface for rotation. Choses the correct rotation.
	   If tree is the root returns false, otherwise - true. */
		 if (tree.father == null)
			 return ( false );
		 if (tree.father.father == null)
			 if(tree.father.left == tree)
				 ZigLeft(tree);
			 else  	ZigRight(tree);
		 else  if(tree.father.left == tree)
				 if(tree.father.father.left == tree.father)
					ZigZigLeft(tree);
			     else  	ZigZagRight(tree);
			   else if(tree.father.father.left == tree.father)
 						ZigZagLeft(tree);
					else ZigZigRight(tree);
		 return (true);
	 }
}




		




    